#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <getopt.h>

#define CHOCOMAIL_VERSION_STRING	"0.0.2"
#define CHOCOMAIL_MAJOR			0
#define CHOCOMAIL_MINOR			0
#define CHOCOMAIL_SUBMINOR		2
#define CHOCOMAIL_VERSION		((CHOCOMAIL_MAJOR<<16)\
					|(CHOCOMAIL_MINOR<<8)\
    					|(CHOCOMAIL_SUBMINOR))

#define ULONG_SIZE 			48
#define URANDOM_DEVICE			"/dev/urandom"
#define RANDOM_PREFIX_DEFAULT_SIZE	5
#define RANDOM_PREFIX_MAX_SIZE		31
#define CRC_DEFAULT_SIZE		3
#define VALID_UNTIL_DAYS		10
#define EMAIL_MAX_SIZE			256

/*
 * Prefix used to redirect the mail to the user
 * in postfix, if the recipient_delimiter is set to +, set prefix to 'loginname+'
 *             if the recipient_delimiter is set to ., set prefix to 'loginname.'
 */
static char *prefix="luc.";
static char *domain="saillard.org";

/*
 * String use to tag a specific address, need to be null terminated.
 * if len(random_prefix)==0, no prefix add
 */
static char   random_prefix[RANDOM_PREFIX_MAX_SIZE+1];
static size_t random_size;	/* Set to 0 by ELF loader */
/*
 * String that contains the crc data, use to protect against spoofing
 * need to be null terminated
 */
static char crc_data[ULONG_SIZE];
static size_t crc_size=CRC_DEFAULT_SIZE;

/*
 * String to indicate the date we want (only six characters)
 * format: YYMMDD
 */
static char expired_date[6];

/*
 * Character used to separate randomprefix,data,and crc
 */
static unsigned char separator='-';

/*
 * The user can choose between and end date or a count reference
 * reference_counter count the number of times a email can be received
 */
static unsigned int reference_counter;	/* Set to 0 by ELF loader */

/*
 *
 *
 *
 */




static unsigned int get_crc(const unsigned char *mem, size_t size, char *out);
static size_t long2hex(unsigned int value, unsigned char *output);
static unsigned int dec2long(const unsigned char *str);
static unsigned int hashpjw(const unsigned char *mem, size_t size);
static int gen_random_prefix(unsigned char *out, size_t size);
static int get_current_date(unsigned char *out, int days_add);
static int gen_email(void);
static int verify_email(const char *email);
static void usage(int exit_code) __attribute__((noreturn));


static struct option chocomail_options[] = {
    /* name       has_arg,  flag,  val */
    {"generate",  0,        0,   'g'},
    {"check",     1,        0,   'c'},
    {"query",     1,        0,   'q'},
    {"help",      0,        0,   'h'},
    {"version",   0,        0,   'v'},

    {"end",       1,        0,   'e'},
    {"domain",    1,        0,   'd'},
    {"tag",       1,        0,   't'},
    {"crc",       1,        0,   's'},
    {"count",     1,        0,     1},
    {0, 0, 0, 0}
};

int main(int argc, char **argv)
{
  int i, opt, option_index;
  int generate_email=1;
  const char *query_email=NULL;
  
  while (1)
   {
     opt = getopt_long (argc, argv, "gc:q:hve:d:t:c:s:",
	                chocomail_options, &option_index);
     if (opt == -1)
       break;

     switch (opt)
      {
       case 'h':
	 usage(0);

       case 'v':
	 printf("chocomail version "CHOCOMAIL_VERSION_STRING"\n"\
	     "Copyright (C) 2003-2004 - Luc Saillard\n"\
	     "This program may be freely redistributed under the terms of the GNU GPL\n");
	 exit(0);

       case 'g':
	 generate_email=1;
	 break;

       case 'c':
       case 'q':
	 query_email = optarg;
	 break;

	 /* **** */
       case 'e':
	if (strlen(optarg)!=6)
	 {
	   fprintf(stderr,"date ending format is invalid\n");
	   exit(1);
	 }
	/* FIXME: check if the date is valid */
	memcpy(expired_date,optarg,6);
	break;

       case 'd':
	domain = optarg;
	break;
	
       case 'p':
	prefix = optarg;
	break;

       case 't':
	random_size = strlen(optarg);
	if (random_size>RANDOM_PREFIX_MAX_SIZE)
	  random_size=RANDOM_PREFIX_MAX_SIZE;
	memcpy(random_prefix,argv[i],random_size);
	random_prefix[random_size] = 0;
	break;

       case 's':
	crc_size = dec2long(optarg);
	break;

       case 1:
	reference_counter = dec2long(optarg);
	break;

	


      }
   }

  if (optind < argc)
   {
     if (*argv[optind]=='+')
      { /* relative date */
	int delay = dec2long(argv[optind]+1);
	if (get_current_date(expired_date,delay)<0)
	 {
	   fprintf(stderr,"date ending format is invalid (with delay)\n");
	   exit(1);
	 }
      }
   }

  /* Check the crc size */
  if (crc_size>7)
    crc_size=7;


  if (query_email)
   {
     return verify_email(query_email);
   } 
  else if (generate_email)
   {
     return gen_email();
   }

  exit(0);
}

/*
 * Print a mini manual and exit
 */
static void usage(int exit_code)
{
  printf("chocomail version "CHOCOMAIL_VERSION_STRING"\n"\
         "Copyright (C) 2003-2004 - Luc Saillard\n"\
	 "This program may be freely redistributed under the terms of the GNU GPL\n"\
	 "\n"\
	 "Usage: chocomail [options]\n"\
	 "\n"\
	 "Options:\n"\
	 " -g, --generate      Generate a new email adress (default)\n"\
	 " -c, --check         Check if the email adress is valid\n"\
	 " -q, --query=EMAIL   Check if the email adress is valid\n"\
	 " -h, --help          Print this usage and exit\n"\
	 " -v, --version       Print the program version and exit\n"\
	 "\n"\
	 "Options for generate mode:\n"\
	 " -e, --end=DATE      Give the last time this email will be valid\n"\
	 "                     DATE is YYMMDD\n"\
	 " -d, --domain=DOMAIN This domain will be put after the @\n"\
	 " -t, --tag=TAG       This is use to identify an unique email adress\n"\
	 " -s, --crc=SIZE      Len of the crc use to verify if an email is not forged\n"\
	 "     --count=COUNT   This adress will not have a end date but a count reference\n"\
	 "\n");
  exit(exit_code);
}


static int gen_email(void)
{
  int i;
  unsigned char *p, *s;
  unsigned char email[EMAIL_MAX_SIZE];

  /* Get the random string */
  if (random_size==0)
   {
     if (gen_random_prefix(random_prefix,RANDOM_PREFIX_DEFAULT_SIZE)<0)
      {
	fprintf(stderr,"Couldn't retrieve some random data\n");
	exit(EX_TEMPFAIL);
      }
     random_size=RANDOM_PREFIX_DEFAULT_SIZE;
   }

  /* Get the expired date */
  if (expired_date[0]==0)
   {
     if (get_current_date(expired_date,VALID_UNTIL_DAYS)<0)
      {
	fprintf(stderr,"Couldn't retrieve the current date\n");
	exit(EX_TEMPFAIL);
      }
   }


  if (  strlen(prefix)
      + random_size
      + 1 /* separator    */
      + 6 /* expired date */
      + 1 /* separator    */
      + 8 /* crc_size */
      + 1 /* @    */
      + strlen(domain)
      + 5 /* dummy value for security */
      > sizeof(email) )
   {
     fprintf(stderr,"Error: email address will be too long\n");
     return -1;
   }

  /*
   * Copy the email address without crc,and domain
   */
  p=email;
  s=prefix;
  while (*s)
    *p++ = *s++;
  s=random_prefix;
  while (*s)
    *p++ = *s++;
  *p++ = separator;

  for (i=0; i<6; i++)
    *p++ = expired_date[i];
  *p++ = separator;
   
  /*
   * Append the crc, CRC is calculated with the prefix/randomstring/date
   */
  get_crc(email,p-email,p);
  /*
   * Append the domain and print the email address
   */
  p+=crc_size;
  *p++='@';
  s=domain;
  while (*s)
    *p++ = *s++;
  *p++=0;

  printf("%s\n",email);
  
}



/**
 * Verify if the address passed in parameter is valid
 * @email: pointer to a string
 * return:
 *   0  no problem
 *   1 email address is outdated (return user unknow)
 *   2 email address is malformed (temp error)
 *
 */
static int verify_email(const char *email)
{
  unsigned char tempbuf[EMAIL_MAX_SIZE+2];
  int i;
  const char *e, *p;
  unsigned char *t;

  if (strlen(email)+crc_size+32 >= EMAIL_MAX_SIZE)
    return 2;

  /* 1st check: verify if we found the prefix */
  p=prefix; e=email; t=tempbuf;
  while (*p)
   {
     if ((*p != *e) || (*e==0))
       return 2; /* go out, if the find the end of email, or character differ */
     *t++ = *e++;
     p++;
   }

  /* 2nd check: this is random prefix (only contains alpha character) */
  if (random_size==0)
    random_size = RANDOM_PREFIX_DEFAULT_SIZE;
  for (i=0;i<random_size;i++)
   {
     if (*e == 0 || *e<'0' || *e>'z')
       return 2; /* go out, string is to short */
     *t++ = *e++;
   }
  /* 3rd: check if we found the separator */
  if (*e != separator)
    return 2;
  *t++ = *e++;

  /* 4th: we have 6 bytes, that need to be a date */
  for (i=0;i<6;i++)
   {
     if (*e<'0' || *e>'9') /* if the email is terminated by \0, it's less than '0' so one test */
       return 2; /* go out, date is only in decimal */
     *t++ = *e++;
   }

  /* 5th: check if we found the separator */
  if (*e != separator)
    return 2;
  *t++ = *e++;

  /*
   * Append the crc, CRC is calculated with the prefix/randomstring/date
   */
  get_crc(tempbuf,t-tempbuf,t);

  //t[crc_size]=0;

  /* 6th: now compare the crc , with the crc found in the email address */ 
  for (i=0;i<crc_size;i++,t++,e++)
   {
     if (*t != *e)
       return 1;
     if (*e==0)
       return 2;
   }

  /* 7th: check if we found @ */
  if (*e != '@')
    return 2;
  
#if 0
  *t++=0;
  printf("Check verify == %s\n",tempbuf);
#endif

  return 0;
}


/**
 * Choose an random prefix
 * @out: buffer in output
 * @size: size of the prefix
 */
static int gen_random_prefix(unsigned char *out, size_t size)
{
  int f;
  unsigned char c;

  f=open(URANDOM_DEVICE,O_RDONLY);
  if (f<0)
    return -1;
  while (size>0)
   {
     if (read(f,&c,1)!=1)
      {
	close(f);
	return -1;
      }
     size--;
     out[size] = (c%26)+'a';
   }
  close(f);
  return 0;
}



/**
 * Generate an hash for a buffer, and return a string
 *
 * @mem:  buffer in input
 * @size: size of the buffer in input
 * @out:  size of a ULONG_SIZE buffer where to put the crc
 * return the number of character written in @out
 */
static unsigned int get_crc(const unsigned char *mem, size_t size, char *out)
{
  return long2hex(hashpjw(mem,size),out);
}

/**
 * Return the date with the format: YYMMDD
 * @out: is a 6 bytes buffer (min)
 * @days_add: add this number of days for the current date
 *
 */
#define todec(x,buf) do { if (x) { *(buf)++=((x)/10)+'0'; *(buf)++=((x)%10)+'0'; } \
   else { *(buf)++='0'; *(buf)++='0'; } } while(0)
static int get_current_date(unsigned char *out, int days_add)
{
  time_t current_time;
  struct tm *current_localtime;
  int r;

  if (time(&current_time) == (time_t)-1)
    return -1;
  current_time+=(days_add*3600*24);	/* convert days in seconds */
  current_localtime = localtime(&current_time);
  if (current_localtime==NULL)
    return -1;

  r=current_localtime->tm_year%100;
  todec(r,out);
  r=current_localtime->tm_mon+1;
  todec(r,out);
  r=current_localtime->tm_mday;
  todec(r,out);

  return 0;
}

/**
 * Convert a unsigned int value into a hexadecimal string
 * (writing all the 32 bits numbers)
 * 
 * @value: value to be converted
 * @output: buffer need to be at least ULONG_SIZE (48)
 * 
 */
#define tohex(x) ((x)>=10?(x)-10+'a':(x)+'0')
static size_t long2hex(unsigned int value, unsigned char *output)
{ 
  int i;
  output+=7;
  for (i=0;i<7;i++)
   {
     *--output = tohex(value&15);
     value>>=4;
   }
  return 8;
}


/**
 * Read a string decimal number and return the number, stop
 * when finding a non decimal character
 *
 * @str: buffer (need to be null terminated)
 */
static unsigned int dec2long(const unsigned char *str)
{
  unsigned char c;	/* Keep this unsigned */
  int sum=0;

  while ((c=*str-'0')<10)
   {
     sum=sum*10+c;
     str++;
   }
  return sum;
}


/**
 * Defines the so called `hashpjw' function by P.J. Weinberger 
 */
static unsigned int hashpjw(const unsigned char *mem, size_t size)
{
  unsigned int g,h=0;
  while (--size)
   {
     h = (h << 4)+ *mem++;
     g = h&0xf0000000;
     if(g)
      {
        h ^= g>>24;
        h ^= g;
      }
   }
  return h;
}




